from base64 import standard_b64encode
import sys

import rich
from rich.console import RenderGroup
from rich.markdown import Markdown
from rich.panel import Panel
from rich.padding import Padding
from rich.syntax import Syntax
from rich.text import Text

from nbtui import _METADATA
from nbtui.parser import (TextCell, MDCell, CodeCell, TextOutputCell,
                    DisplayOutputCell, BlankCell, ErrorOutputCell)

_RULE = rich.rule.Rule(style="white", end="")

class Notebook:
    def __init__(self, cells):
        self.cell_displays = {}
        self.cell_renders = {}
        self.plots_todraw = []

        # which line is currently at the top
        self.row = 0
        self.search_pat = None
        self.needs_redraw = False

        display_row = 0

        for cell in cells:
            self.cell_displays[display_row] = cell
            display_row += cell.n_lines


        # This is a bit ugly: if the notebook is too small, rich.Live will
        # not hide the shell prompt, which throws off calculations
        # of display offsets.

        # Therefore, add an artificial empty cell to ensure that
        # the size of the notebook exceeds the terminal height.

        if display_row < _METADATA["term_height"]:
            dummy_text = ["\n" for _ in range(_METADATA["term_height"] - 
                                              display_row + 1)]
            dummy_cell = CodeCell(dummy_text)
            self.cell_displays[display_row] = dummy_cell
            self.cell_renders[display_row] = None

        self.size = max(display_row, _METADATA["term_height"])

    def draw_plot_later(self, cell, start_row):
        self.plots_todraw.append((cell.b64,
            (start_row, int((_METADATA["term_width"] - cell.size[1]) / 2)),
            cell.size))

    def draw_plots(self):
        for (image, pos, size) in self.plots_todraw:
            display_image(image, pos, size)

        self.plots_todraw.clear()

    def get_renders_in_range(self, start, end):
        """
        returns all renders between {start} and {end},
        truncating any starting cells as necessary.
        """
        renders = []
        for k, v in self.cell_displays.items():

            if k + v.n_lines > start and k < start:
                # only part of the cell is showing
                truncated_cell = v.truncate(start - k)

                renderable = render_cell(truncated_cell)

                if truncated_cell.pad:
                    renderable = pad_renderable(renderable, start - k)

                renders.append(renderable)

                if isinstance(truncated_cell, DisplayOutputCell):
                    self.draw_plot_later(truncated_cell,
                            max(3, 5 - (start - k)))

            elif k >= start and k <= end:
                if self.cell_renders.get(k, None) is None:
                    self.cell_renders[k] = pad_renderable(
                            render_cell(self.cell_displays[k]), 0)

                renders.append(self.cell_renders[k])

                if isinstance(v, DisplayOutputCell) and k + v.n_lines <= end:
                    self.draw_plot_later(v, k - start + 5)
                elif isinstance(v, DisplayOutputCell) and k + v.n_lines > end:
                    # Normally, in the case of bottom truncation, we push the
                    # entire renderable on, and let Rich handle the cropping.
                    # However, since plot drawing is done separately,
                    # we have to handle bottom truncation of images ourselves.
                    truncated_cell = v.truncate_bottom(end - k)

                    if isinstance(truncated_cell, DisplayOutputCell):
                        self.draw_plot_later(truncated_cell, 
                                _METADATA["term_height"] - (end - k - 5))

        return renders


def display_notebook(notebook):
    row = notebook.row
    renders = notebook.get_renders_in_range(row, row + _METADATA["term_height"])

    notebook.needs_redraw = False

    return Panel(RenderGroup(*renders))

def render_cell(cell):
    if isinstance(cell, BlankCell):
        return Syntax(" \n" * cell.n_lines, "python",
                      background_color="default")
    elif isinstance(cell, MDCell):
        return Markdown(cell.text)
    elif isinstance(cell, CodeCell):
        return Syntax(cell.text, _METADATA["language"],
                background_color="default")
    elif isinstance(cell, TextOutputCell):
        return Text(cell.text)
    elif isinstance(cell, DisplayOutputCell):
        # draw a large blank space for the plot to fit into
        return Syntax(" \n" * (cell.n_lines - 3), "python",
            background_color="default")
    elif isinstance(cell, ErrorOutputCell):
        return Text.from_markup(cell.tb_text)
    else:
        assert False


def pad_renderable(renderable, offset):
    """
    Pad a renderable, subject to a particular truncation offset.
    """
    if offset < 0:
        raise Exception("invalid offset!")
    if offset == 0:
        return RenderGroup(_RULE, Padding(renderable, 1))
    if offset == 1:
        return Padding(renderable, 1)
    else:
        return Padding(renderable, (0, 1, 1, 1))

def display_image(image, position, size):
    """
    Takes a base 64 encoded image string, and displays it on the screen
    using the kitty graphics protocol.

    image: base 64 encoded png
    """
    # move cursor
    sys.stdout.buffer.write(b'\033[%d;%dH' % position)

    cmd = {"a": "T", "f": 100,  "r": size[0], "c":size[1]}

    while image:
        chunk, image = image[:4096], image[4096:]

        m = 1 if image else 0
        cmd["m"] = m

        cmd_header = ','.join(f'{k}={v}' for k, v in cmd.items())
        cmd_header = cmd_header.encode("ascii")
        payload = b''.join((b'\033_G', cmd_header, b';', chunk, b'\033\\'))

        sys.stdout.buffer.write(payload)
        sys.stdout.flush()

        cmd.clear()



from contextlib import contextmanager
from functools import partial
import os
import re
import signal
import sys
import termios

from nbtui import _METADATA
from nbtui.parser import TextCell
from nbtui.display import display_notebook

class SetTermAttrs:
    def __init__(self, fd):
        self.fd = fd

    def __enter__(self):
        self.oldattr = termios.tcgetattr(self.fd)
        newattr = termios.tcgetattr(self.fd)

        # Disable CANONICAL and ECHO modes for stdin lflags, so 
        # that input becomes available immediately, instaed of after
        # newline
        newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(self.fd, termios.TCSANOW, newattr)

        signal.signal(signal.SIGINT, self._sigint_handler)

    def __exit__(self, type, value, traceback):
        termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.oldattr)

    def _sigint_handler(self, signal_recv, frame):
        self.__exit__(None, None, None)

        sys.exit(0)

def get_char():
    c = sys.stdin.read(1)
    return c if c else ""

def scroll(n, notebook):
    if notebook.row + n <= 0:
        notebook.row = 0
    # the + 2 is for the two extra rows from rich.Panel
    elif notebook.row + n > notebook.size + 2 - _METADATA["term_height"]:
        notebook.row = notebook.size + 2 - _METADATA["term_height"]
    else:
        notebook.row += n
    return False

def goto(row, notebook):
    # clamp to end of notebook
    row = min(row, notebook.size + 2 - _METADATA["term_height"])

    # negative values indicate end of notebook
    if row < 0:
        row = notebook.size + 2 - _METADATA["term_height"]

    notebook.row = row
    return False

def search(forward, notebook):
    # place cursor at bottom of screen
    if forward == True:
        sys.stdout.buffer.write(b'\033[999;1H/')
    else:
        sys.stdout.buffer.write(b'\033[999;1H?')

    search_pat = ".*?" + input("")
    search_pat = re.compile(search_pat)

    notebook.search_pat = search_pat

    if forward == True:
        return search_next(notebook)
    else:
        return search_prev(notebook)

def search_next(notebook):
    if notebook.search_pat is None:
        return False

    for line, cell in notebook.cell_displays.items():
        if line + cell.n_lines < notebook.row:
            continue

        if not isinstance(cell, TextCell):
            continue

        try:
            offset = next(i for i, l in enumerate(cell.text_lines) if
                          (notebook.search_pat.match(l) and
                              i + line > notebook.row))
            goto(line + offset + 2, notebook)
            return False
        except StopIteration:
            continue

    return False

def search_prev(notebook):
    if notebook.search_pat is None:
        return False

    for line, cell in reversed(notebook.cell_displays.items()):
        # Skip to the next if we are on the first line of a cell
        if line > notebook.row - 3:
            continue

        if not isinstance(cell, TextCell):
            continue

        try:
            offset = next(i for i, l in enumerate(reversed(cell.text_lines)) if
                          (notebook.search_pat.match(l) and
                              line + i < notebook.row))
            goto(line + offset + 2, notebook)
            return False
        except StopIteration:
            continue

    return False

def exit(_):
    return True

input_dict = {
            "j": partial(scroll, 1),
            "k": partial(scroll, -1),
            '\x04': partial(scroll, 15), # CTRL-D
            '\x15': partial(scroll, -15), # CTRL-U
            "G": partial(goto, -1),
            "g": partial(goto, 0),
            "/": partial(search, True),
            "?": partial(search, False),
            "n": search_next,
            "N": search_prev,
            'q': exit,
        }

def handle_input(char, notebook):
    if char == "":
        return False
    try:
        stop = input_dict[char](notebook)
        notebook.needs_redraw = True
        return stop
    except KeyError:
        return False



from base64 import decodebytes, b64encode
from dataclasses import dataclass
import io
from math import ceil, floor
import logging
import re

from PIL import Image

from nbtui import _METADATA
import nbtui.display

def parse_nb(json_notebook):
    """
    Parses a json representation of a jupyter notebook,
    and returns a list of cells and their outputs, thus representing
    the parsed notebook.
    """
    json_notebook = json_notebook["cells"]
    parsed_cells = []
    for cell in json_notebook:
        parsed_cells.append(parse_nb_cell(cell))
        # break up cells and outputs into distinct units
        if cell.get("outputs", None) is not None:
            for output in cell["outputs"]:
                output_cell = parse_nb_output(output)
                if output_cell is not None:
                    parsed_cells.append(output_cell)

    return parsed_cells

def parse_nb_cell(cell):
    if cell["cell_type"] == "markdown":
        return MDCell(cell["source"])
    
    return CodeCell(cell["source"])

def parse_nb_output(output):
    if output["output_type"] == "stream":
        return CodeCell(output["text"])
    elif output["output_type"] == "error":
        return ErrorOutputCell(output["traceback"])
    elif output["output_type"] == "display_data":
        im = try_parse_image(output)
        if im is not None:
            return DisplayOutputCell(im, "png")

        # if there is no png, just return a blank line
        return BlankCell(1)
    elif output["output_type"] == "execute_result":
        im = try_parse_image(output)
        if im is not None:
            return DisplayOutputCell(im, "png")

        json = output["data"].get("application/json", None)
        if json is not None:
            return CodeCell([str(json)])

        # if we didn't find anything else, return the text
        text = output["data"].get("text/plain", None)
        if text is not None:
            return TextOutputCell(text)
    raise Exception(
            f"Encountered unparsable cell output type {output['output_type']}"
            )

def reparse_nb(json_notebook, parsed_notebook):
    """
    Given a modified version of the notebook,
    reparse the notebook and update any changes.
    For now, re-render everything on change
    """
    num_cells = 0

    for cell in json_notebook["cells"]:
        num_cells += 1
        try:
            num_cells += len(cell["outputs"])
        except:
            pass

    if num_cells != len(parsed_notebook.cell_displays):
        # TODO: Can we avoid reparsing the entire notebook
        # whenever a cell is added/deleted?
        new_nb = nbtui.display.Notebook(parse_nb(json_notebook))
        new_nb.needs_redraw = True
        return new_nb

    json_notebook = json_notebook["cells"]

    parsed_cells = list(parsed_notebook.cell_displays.items())

    offset = 0
    new_cells = {}

    for cell in json_notebook:
        display_line, old_cell = parsed_cells.pop(0)
        if old_cell.compare(cell):
            new_cells[display_line + offset] = old_cell
        else:
            new_cell = parse_nb_cell(cell)
            offset += new_cell.n_lines - old_cell.n_lines
            new_display_line = offset + display_line
            # The display row of the first cell doesn't change
            if display_line == 0:
                new_display_line = 0
            new_cells[new_display_line] = new_cell

        if cell.get("outputs", None) is not None:
            for output in cell["outputs"]:

                display_line, old_output = parsed_cells.pop(0)
                # breakpoint()
                if old_output.compare(output):
                    new_cells[display_line + offset] = old_output
                else:
                    new_output = parse_nb_output(output)
                    offset += new_output.n_lines - old_output.n_lines
                    new_cells[display_line + offset] = new_output

    parsed_notebook.cell_displays = new_cells
    parsed_notebook.cell_renders = {}
    parsed_notebook.needs_redraw = True

    return parsed_notebook

def try_parse_image(cell):
    if not _METADATA["img_support"]:
        return None

    return cell["data"].get("image/png", None)

class BlankCell:
    pad = False

    def __init__(self, n):
        self.n_lines = n

    def register_draw(self, notebook, offset):
        pass

    def compare(self, cell):
        return True

class TextCell:
    pad = True

    def __init__(self, text):
        self.for_compare = hash("".join(text))

        self.n_lines = len(text) + 3
        # to ensure blank lines get rendered correctly,
        # replace a blank line with a single space
        self.text_lines = text
        self.text = "".join((t if t != "\n" else " \n" for t in text))

    def truncate(self, offset):
        if offset >= self.n_lines - 1:
            return BlankCell(1)
        idx = max(offset - 2, 0)
        truncated_lines = self.text_lines[idx:]
        return type(self)(truncated_lines)

    def compare(self, cell):
        return self.for_compare == hash(TextCell.get_text_from_json(cell))

    @staticmethod
    def get_text_from_json(json_cell):
        if json_cell.get("cell_type", None) in ("code", "markdown"):
            return "".join(json_cell["source"])

        # must be an output cell
        if json_cell["output_type"] == "stream":
            return "".join(json_cell["text"])

        # must be execute_result
        json = json_cell["data"].get("application/json", None)
        if json is not None:
            return str(json)

        return "".join(json_cell["data"]["text/plain"])

class MDCell(TextCell):
    pass

class CodeCell(TextCell):
    pass

class TextOutputCell(TextCell):
    pass

@dataclass
class ErrorOutputCell:
    ANSI_COLOR_DICT = {
            "\x1b[0;31m": "[red]",
            "\x1b[0;32m": "[green]",
            "\x1b[0;33m": "[yellow]",
            "\x1b[0;34m": "[blue]",
            "\x1b[0;35m": "[magenta]",
            "\x1b[0;36m": "[cyan]",
            "\x1b[0;37m": "[white]",
            "\x1b[1;31m": "[bold red]",
            "\x1b[1;32m": "[bold green]",
            "\x1b[1;33m": "[bold yellow]",
            "\x1b[1;34m": "[bold blue]",
            "\x1b[1;35m": "[bold magenta]",
            "\x1b[1;36m": "[bold cyan]",
            "\x1b[1;37m": "[bold white]",
            "\x1b[0m":    "[/]"
            }
    pad = True

    def __init__(self, traceback, needs_processing=True):
        self.for_compare = hash("".join(traceback))

        self.traceback = []
        if needs_processing:
            # Jupyter notebook traceback comes with a bunch of ansi
            # color escape codes. We need to convert these to rich
            # markup, in order for these to be displayed correctly.
            for entry in traceback:
                entry = entry.replace("-", "─")
                entry = entry.replace("─>", "─→")
                for color, markup in self.ANSI_COLOR_DICT.items():
                    entry = entry.replace(color, markup)

                rows = entry.split("\n")
                for row in rows:
                    row = self.fix_markup(row)
                    self.traceback.append(row if row != "\n" else " \n")
        else:
            self.traceback = traceback

        self.tb_text = "\n".join(self.traceback)
        self.n_lines = len(self.traceback) + 3

    def truncate(self, offset):
        idx = max(offset - 2, 0)
        return ErrorOutputCell(self.traceback[idx:], needs_processing=False)

    def compare(self, other):
        other_text = "".join(other["traceback"])
        return self.for_compare == hash(other_text)

    @staticmethod
    def fix_markup(text):
        n_style = 0
        idx = 0
        while True:
            pos = re.search("\[(.*?)\]", text[idx:])
            if not pos:
                break

            if pos.group(1) == "/":
                n_style -= 1
            else:
                n_style += 1

            if n_style < 0:
                text = text[:idx + pos.span()[0]] + text[idx + pos.span()[1]:]
                n_style = 0
            else:
                idx += pos.span()[1]

        return text

class DisplayOutputCell:
    def __init__(self, b64_data, fmt):
        self.for_compare = hash(b64_data)

        # strip off the final newline
        b64 = b64_data.encode("ascii")[:-1]
        self.b64 = b64
        self.fmt = fmt

        # note - PIL sizes are (width x height)
        self.img = Image.open(io.BytesIO(decodebytes(b64)))
        width = self.img.size[0]
        height = self.img.size[1]

        if (width >= (_METADATA["term_width"] / 1.5) *
                _METADATA["pix_per_col"] or 
            height >= (_METADATA["term_height"] / 1.5) *
                _METADATA["pix_per_row"]):

            width = min(width, floor((_METADATA["term_width"] / 1.5) *
                                     _METADATA["pix_per_col"]))
            height = min(height, floor((_METADATA["term_height"] / 1.5) *
                                       _METADATA["pix_per_row"]))
            self.img = self.img.resize((width, height))
            self.b64 = self.img_to_b64(self.img)

        self.size = (ceil(height / _METADATA["pix_per_row"]),
                     ceil(width / _METADATA["pix_per_col"]))
        self.n_lines = self.size[0] + 5
        self.pad = True

    def truncate(self, offset):
        if offset <= 2:
            return self
        if offset >= self.n_lines - 4:
            return BlankCell(self.n_lines - offset)

        offset_in_px = ceil((offset - 2) * _METADATA["pix_per_row"])
        new_img = self.img.crop((0, offset_in_px,
            self.img.size[0], self.img.size[1]))

        new_data = DisplayOutputCell.img_to_b64(new_img).decode("utf-8") + "\n"
        return DisplayOutputCell(new_data, self.fmt)

    def truncate_bottom(self, offset):
        if offset <= 4:
            return BlankCell(1)
        
        offset_in_px = ceil((offset - 4) * _METADATA["pix_per_row"])

        new_img = self.img.crop((0, 0, self.img.size[0],
            offset_in_px))

        new_data = DisplayOutputCell.img_to_b64(new_img).decode("utf-8") + "\n"
        return DisplayOutputCell(new_data, self.fmt)

    def compare(self, other):
        return self.for_compare == hash(other["data"]["image/png"])

    @staticmethod
    def img_to_b64(img):
        stream = io.BytesIO()
        img.save(stream, format="png")
        return b64encode(stream.getvalue())

if __name__ == "__main__":
    out = ErrorOutputCell.fix_markup("[red]hello [/][blue] world [/][/]")



import setuptools

with open("README.md", "r") as f:
    desc = f.read()

requirements = []
with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="nbtui",
    version="0.0.2",
    author="Tony Chen",
    author_email="tchen1998@gmail.com",
    description="View jupyter notebooks (with images) from the command line",
    long_description=desc,
    long_description_content_type="text/markdown",
    url="https://github.com/chentau/nbtui",
    packages=setuptools.find_packages(),
    entry_points={
            "console_scripts": [
                    "nbtui=nbtui.__main__:main"
                ]
        },
    python_requires='>=3.6',
    install_requires = requirements
)



#!/bin/python

import argparse
import array
import fcntl
import json
import multiprocessing as mp
import sys
import termios
import os

import rich
from rich.live import Live
from watchgod import run_process, RegExpWatcher

from nbtui import _METADATA
from nbtui.display import display_notebook, Notebook
from nbtui.parser import parse_nb, reparse_nb
from nbtui.user_input import SetTermAttrs, get_char, handle_input

def check_resized():
    term_width, term_height= os.get_terminal_size()
    return (term_width != _METADATA.get("term_width", None) or
            term_height != _METADATA.get("term_height", None))

def parse_metadata():
    term_width, term_height= os.get_terminal_size()

    buf = array.array('H', [0, 0, 0, 0])
    fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, buf)
    screen_width = buf[2]
    screen_height = buf[3]

    _METADATA["img_support"] = False
    if screen_height != 0 and screen_width != 0:
        _METADATA["img_support"] = True

    pixels_per_row = screen_height / term_height
    pixels_per_col = screen_width / term_width

    _METADATA["term_height"] = term_height
    _METADATA["term_width"] = term_width
    _METADATA["screen_width"] = buf[2]
    _METADATA["screen_height"] = buf[3]
    _METADATA["pix_per_row"] = pixels_per_row
    _METADATA["pix_per_col"] = pixels_per_col

def filewatch_worker(queue, filename):

    def on_changed(queue, filename):
        with open(filename, "r") as f:
            new_nb = json.load(f)
        queue.put(new_nb)

    sys.stderr = open(os.devnull, "w")
    sys.stdout = open(os.devnull, "w")

    filename = os.path.realpath(filename)
    start_dir = os.path.dirname(filename)
    run_process(start_dir, on_changed, watcher_cls = RegExpWatcher,
            watcher_kwargs = {"re_files": filename},
            args=(queue, filename))

def input_worker(in_queue, out_queue):
    sys.stdin = open(0)
    fd = sys.stdin.fileno()
    while True:
        with SetTermAttrs(fd):
            c = get_char()
            in_queue.put(c)

        # Block until main proc has processed the keypress
        # this is so we don't clobber stdin for things like searching,
        # where the user types additional characters after the inital
        # press of /
        _ = out_queue.get()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", type=str)
    args = parser.parse_args()

    filename = args.filename

    if filename[-6:] != ".ipynb":
        raise Exception("Only accepts jupyter notebooks")
    
    with open(filename, "r") as f:
        nb = json.load(f)

    _METADATA["language"] = nb["metadata"]["kernelspec"]["language"]
    parse_metadata()
    notebook = Notebook(parse_nb(nb))

    filewatch_queue = mp.Queue()
    filewatch_p = mp.Process(target=filewatch_worker,
                             args=(filewatch_queue, filename))
    filewatch_p.start()

    input_queue = mp.Queue()
    out_queue = mp.Queue()
    input_p = mp.Process(target=input_worker, daemon=True,
                         args=(input_queue, out_queue))
    input_p.start()

    rendered_cells = display_notebook(notebook)

    with Live(transient=True,
              auto_refresh=False,
              vertical_overflow="crop",
              redirect_stdout=False) as live:

        live.update(rendered_cells, refresh=True)
        notebook.draw_plots()
        char = ''
        stop = False
        while not stop:
            if notebook.needs_redraw:
                rendered_cells = display_notebook(notebook)

                sys.stdout.buffer.write(b"\x1b[2J\x1b[H")
                live.update(rendered_cells, refresh=True)
                notebook.draw_plots()

            if not filewatch_queue.empty():
                new_nb = filewatch_queue.get()
                notebook = reparse_nb(new_nb, notebook)

            if not input_queue.empty():
                char = input_queue.get()
                stop = handle_input(char, notebook)
                # signal that the keypress was handled
                out_queue.put(1)

            if check_resized():
                parse_metadata()
                notebook.needs_redraw = True

    filewatch_queue.close()
    input_queue.close()

    # input subproc will be automatically closed, since it was initialized
    # as a daemon
    filewatch_p.terminate()
    filewatch_p.join()
    sys.stdout.buffer.write(b"\x1b[2J\x1b[H")

if __name__ == "__main__":
    main()
